#include <math.h>
#include <qapplication.h>
#include <qsplitter.h>
#include <qlayout.h>
#include <qpushbutton.h>
#include <qpainter.h>
#include <qprogressdialog.h>
#include <qurl.h>

#ifdef QT4
#include <QDesktopWidget>
#include <QHeaderView>
#ifdef MAEMO
#include <QWindowsStyle>
#endif
#endif

#ifdef QTOPIA
#include <qpe/qpemenubar.h>
#else
#include <qmenubar.h>
// for Qtopia we are using the QPE variants for toolbar and menubar
// to enable special features like the menu key etc.
// Define aliases to avoid ifdefs...
#define QPEMenuBar QMenuBar
#define QPEToolBar QToolBar
#endif

#include "coverartdownloader.h"
#include "reflectionimageflow.h"
#include "coverartprovider.h"
#include "coverartflow.h"
#include "debug.h"

#include "compathack.h"

#define cCaptionHeight 12
#define cSearchGrabberCount 8
#define cDownloadGrabberCount 2


static CoverArtDownloaderAddOn addOn;


/* DownloadedCoverArtImageFlow */

DownloadedCoverArtImageFlow::DownloadedCoverArtImageFlow(LayeredPaintBox *parent)
    :	CustomPerspectiveImageFlow(parent),
        albumListViewItem_(NULL)
{
#ifndef QT4
    images_.setAutoDelete(true);
#endif

    setTransitionTime(300);
    setConstantVerticalOffset(-cCaptionHeight);
    setSlideSize(QSize(300, 300));
}

DownloadedCoverArtImageFlow::~DownloadedCoverArtImageFlow()
{
#ifdef QT4
    qDeleteAll(images_);
#endif
}

QImage DownloadedCoverArtImageFlow::getImage(int index)
{
#ifdef QT4
    if (index >= images_.count())
        return QImage();
#else
    if (index >= images_.count())
        return NULL;
#endif

    return *images_.at(index);
}

void DownloadedCoverArtImageFlow::setAlbumListViewItem(AlbumListViewItem *albumListViewItem)
{
    albumListViewItem_ = albumListViewItem;

    setItemCount(albumListViewItem_->coverArtSearchResults.count());

    if (albumListViewItem_->selectedResultIndex == -1)
        showIndex(0);
    else
        showIndex(albumListViewItem_->selectedResultIndex);
}

void DownloadedCoverArtImageFlow::paintOverlay(QPainter &p, const QPixmap &pixmap)
{
    CustomPerspectiveImageFlow::paintOverlay(p, pixmap);

    p.setPen(QColor(255, 255, 255));
#ifdef QT4
    p.setBrush(Qt::NoBrush);
#else
    p.setBrush(NoBrush);
#endif

    QRect rect;
    int itemIndex = floor(currentValue());

    if (itemIndex >= 0 && itemIndex < images_.count())
    {
        CoverArtSearchResult *item = albumListViewItem_->coverArtSearchResults.at(itemIndex);

        int boxHeight = cCaptionHeight;
#ifdef QTOPIA
        int fontHeight = 15; // TODO: FIXME!
#else
        QFont font = p.font();
        font.setPixelSize(10);
        p.setFont(font);
        int fontHeight = p.fontInfo().pixelSize();
#endif

        QString resolution;
        QImage *image = images_.at(itemIndex);

        if (image->isNull())
            resolution = "Not loaded yet.";
        else
            resolution = QString::number(image->width()) + " x " + QString::number(image->height()) + " Pixel";

        rect = QRect(0, height() - boxHeight, width(), boxHeight);
        p.drawText(rect, Qt::AlignCenter, resolution + " - " + item->imageURL);
    }

    if (!albumListViewItem_)
    {
        rect = QRect(0, 0, width(), height());
        p.drawText(rect, Qt::AlignCenter, tr("Please select an entry in the list above to review the search results."));
    }
    else if (images_.count() == 0)
    {
        rect = QRect(0, 0, width(), height());
        //p.drawText(rect, Qt::AlignCenter, tr("No albums were found that are missing cover art images."));
        p.drawText(rect, Qt::AlignCenter, tr("No cover art images were found for the selected album."));
    }
}

void DownloadedCoverArtImageFlow::transitionFinished()
{
    DENTERMETHOD;

    CustomPerspectiveImageFlow::transitionFinished();

    int itemIndex = floor(currentValue());
    CoverArtSearchResult *item = albumListViewItem_->coverArtSearchResults.at(itemIndex);
    albumListViewItem_->selectedResultIndex = itemIndex;

    albumListViewItem_->setText(3, QString::number(itemIndex + 1));

    DEXITMETHOD;
}

/* HttpGrabber */

HttpGrabber::HttpGrabber(QObject *parent)
    :	QObject(parent),
        http_(parent),
        url_(""),
        userData_(NULL),
        timeout_(5000),
        busy_(false),
        downloadRes_(0),
        lastBytesAvailable_(0)
{
    connect(&http_, SIGNAL(requestFinished(int, bool)), this, SLOT(httpRequestFinished(int, bool)));
    connect(&timer_, SIGNAL(timeout()), this, SLOT(timerTimeout()));
}

HttpGrabber::~HttpGrabber()
{
    abort();
}

bool HttpGrabber::download(const QString &url, void *userData)
{
    DENTERMETHODF("%p [%s, %p]", &http_, (const char *)url.toUtf8(), userData);

    if (http_.hasPendingRequests())
        http_.abort();

    timer_.stop();
    lastBytesAvailable_ = 0;

#ifdef QT4
    timer_.start(timeout_);
#else
    timer_.start(timeout_, false);
#endif

    busy_ = true;
    url_ = url;
    userData_ = userData;

    QString proxyHost = QUrl(url).host();
    int proxyPort = 80;

#ifndef QT4
    getHTTPProxySettings(proxyHost, proxyPort);
#endif

    http_.setHost(proxyHost, proxyPort);

    downloadRes_ = http_.get(url);

    DPRINTF("downloadRes_: %d", downloadRes_);

    DEXITMETHODF("%p [%s, %p]", &http_, (const char *)url.toUtf8(), userData);
}

void HttpGrabber::abort()
{
    downloadRes_ = 0;
    timer_.stop();
    http_.abort();
    busy_ = false;
}

void HttpGrabber::timerTimeout()
{
    DENTERMETHODF("%p", &http_);

    uint bytesAvailable = http_.bytesAvailable();

    if (bytesAvailable == lastBytesAvailable_)
    {
        abort();
        emit finished(this, url_, userData_, true, true);
    }
    else
        lastBytesAvailable_ = bytesAvailable;

    DEXITMETHODF("%p", &http_);
}

void HttpGrabber::httpRequestFinished(int res, bool error)
{
    DENTERMETHODF("%p [%d, %s]", &http_, res, error ? "true" : "false");

    if (res == downloadRes_)
    {
        timer_.stop();
        busy_ = false;

        if (!error)
        {
            QByteArray data = http_.readAll();
            emit dataAvailable(this, url_, userData_, &data);
            emit finished(this, url_, userData_, false, false);
        }
        else
            emit finished(this, url_, userData_, true, false);
    }

    DEXITMETHODF("%p [%d, %s]", &http_, res, error ? "true" : "false");
}

/* CoverArtDownloader */

CoverArtDownloader::CoverArtDownloader(PlayList *playList, SkinManager* skinManager, QWidget *parent)
    :	QDialog(parent),
        playList_(playList),
        skinManager_(skinManager),
        searchingOnline_(false),
        downloadingImageFiles_(false),
        currentAlbumIndex_(-1),
        currentResultIndex_(-1),
        coverArtSearchResultsToDownload_(NULL)
{
#ifndef QTOPIA
#ifdef QT4
#ifdef MAEMO
    setAttribute(Qt::WA_Maemo5StackedWindow);
    setWindowFlags(Qt::Window);
#else
    setWindowFlags(Qt::Window);
    setAttribute(Qt::WA_DeleteOnClose);
#endif
#else
    setWFlags(WType_TopLevel | WStyle_Customize | WStyle_NormalBorder | WStyle_Title | WStyle_SysMenu | WStyle_MinMax | WDestructiveClose);
#endif
#endif

#ifdef QTOPIA
    setWState(WState_Reserved1);
#endif
    setSizeGripEnabled(true);

    setWindowTitle(tr("Quasar Cover Art Downloader"));

    QVBoxLayout *mainLayout = new QVBoxLayout(this);
    mainLayout->setSpacing(-1);
    mainLayout->setMargin(0);
    QVBoxLayout *contentLayout = new QVBoxLayout(this);
    contentLayout->setSpacing(-1);
    contentLayout->setMargin(0);

    splitter_ = new QSplitter(this);
    splitter_->setOrientation(Qt::Vertical);

#ifdef QT4
    listView_ = new QTreeWidget(splitter_);

    listView_->setAllColumnsShowFocus(true);
    listView_->setColumnCount(5);
    listView_->setHeaderLabels(QStringList()
        << tr("Artist")
        << tr("Album")
        << tr("Results")
        << tr("Image#")
        << tr("Status")
    );

    listView_->setItemsExpandable(false);
    listView_->setUniformRowHeights(true);
    listView_->setRootIsDecorated(false);
    listView_->setWordWrap(false);
    listView_->setAllColumnsShowFocus(true);
    listView_->setAlternatingRowColors(true);
    listView_->setVerticalScrollMode(QAbstractItemView::ScrollPerPixel);
    listView_->header()->setStretchLastSection(false);

    connect(listView_, SIGNAL(currentItemChanged(QTreeWidgetItem*, QTreeWidgetItem*)), this, SLOT(listViewCurrentItemChanged()));
#else
    listView_ = new QListView(splitter_);

    listView_->addColumn(tr("Artist"));
    listView_->addColumn(tr("Album"));
    listView_->addColumn(tr("Results"));
    listView_->addColumn(tr("Image#"));
    listView_->addColumn(tr("Status"));

    connect(listView_, SIGNAL(selectionChanged(QListViewItem *)), this, SLOT(listViewCurrentItemChanged()));
#endif

    imageFlowViewBox_ = new LayeredPaintBox(splitter_);
    connect(imageFlowViewBox_, SIGNAL(rearrangeLayers()), this, SLOT(imageFlowViewBoxRearrangeLayers()));

    imageFlowView_ = new DownloadedCoverArtImageFlow(imageFlowViewBox_);
    flowScrollBar_ = new SkinFlowScrollBar(Skin::FlowScrollBar, Skin::FlowScrollBarSlider, skinManager_, imageFlowView_, imageFlowViewBox_);

    listView_->installEventFilter(this); // Install event filter to intercept left and right key press events...

    contentLayout->addWidget(splitter_);

#ifdef QT4
    QList<int> sizes;
#else
    QValueList<int> sizes;
#endif
    sizes.append(50);
    sizes.append(50);
    splitter_->setSizes(sizes);

#ifndef MAEMO
    QWidget *twoButtonsBox = new QWidget(this);
    twoButtonsBox->setSizePolicy(QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed));

    QHBoxLayout *twoButtonsBoxLayout = new QHBoxLayout(twoButtonsBox);
    twoButtonsBoxLayout->setSpacing(8);
    twoButtonsBoxLayout->setMargin(0);

    startDownloadButton_ = new QPushButton(tr("Start Download"), twoButtonsBox);
    connect(startDownloadButton_, SIGNAL(clicked()), this, SLOT(accept()));
    startDownloadButton_->setSizePolicy(QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed));
    twoButtonsBoxLayout->addWidget(startDownloadButton_);

    cancelButton_ = new QPushButton(tr("Close"), twoButtonsBox);
    connect(cancelButton_, SIGNAL(clicked()), this, SLOT(reject()));
    cancelButton_->setSizePolicy(QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed));
    twoButtonsBoxLayout->addWidget(cancelButton_);
#endif

    progressLabel_ = new QLabel(this);
    progressLabel_->setSizePolicy(QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed));

    progress_ = new QProgressBar(this);

    QHBoxLayout *buttonsLayout = new QHBoxLayout(this);
    buttonsLayout->setSpacing(-1);
    buttonsLayout->setMargin(15);
    buttonsLayout->setDirection(QBoxLayout::RightToLeft);
    buttonsLayout->setAlignment(Qt::AlignRight);

#ifndef MAEMO
    buttonsLayout->addWidget(twoButtonsBox, 1);
#endif
    buttonsLayout->addWidget(progress_, 1);
    buttonsLayout->addWidget(progressLabel_, 1);

#ifdef QT4
    QMenu *tools = new QMenu("&Tools", this);
    tools->addAction(tr("&Refresh list..."), this, SLOT(scanForMissingCoverArt()));
    tools->addAction(tr("&Start Download"), this, SLOT(accept()));
#ifndef MAEMO
    tools->addSeparator();
    tools->addAction(tr("&Close"), this, SLOT(reject()));
#endif

    QMenuBar *menu = new QMenuBar(this);
    menu->addMenu(tools);
    mainLayout->setMenuBar(menu);
#else
    QPopupMenu *tools = new QPopupMenu(this);
    tools->insertItem(tr("&Refresh list..."), this, SLOT(scanForMissingCoverArt()));
    tools->insertItem(tr("&Start Download"), this, SLOT(accept()));
    tools->insertSeparator();
    tools->insertItem(tr("&Close"), this, SLOT(reject()));

    QPEMenuBar *menu = new QPEMenuBar(this);
    menu->setMargin(0);
    menu->insertItem("&Tools", tools);

    mainLayout->addWidget(menu);
#endif

    mainLayout->addLayout(contentLayout);
    mainLayout->addLayout(buttonsLayout);

#ifndef QT4
    searchGrabbers_.setAutoDelete(true);
#endif
    for (int i = 0; i < cSearchGrabberCount; ++i)
    {
        HttpGrabber *grabber = new HttpGrabber(this);
        grabber->setTimeout(30000);
        connect(grabber, SIGNAL(dataAvailable(HttpGrabber *, const QString &, void *, QByteArray *)), this, SLOT(searchDataAvailable(HttpGrabber *, const QString &, void *, QByteArray *)));
        connect(grabber, SIGNAL(finished(HttpGrabber *, const QString &, void *, bool, bool)), this, SLOT(searchFinished(HttpGrabber *, const QString &, void *, bool, bool)));
        searchGrabbers_.append(grabber);
    }

#ifndef QT4
    downloadGrabbers_.setAutoDelete(true);
#endif
    for (int i = 0; i < cDownloadGrabberCount; ++i)
    {
        HttpGrabber *grabber = new HttpGrabber(this);
        connect(grabber, SIGNAL(dataAvailable(HttpGrabber *, const QString &, void *, QByteArray *)), this, SLOT(downloadDataAvailable(HttpGrabber *, const QString &, void *, QByteArray *)));
        connect(grabber, SIGNAL(finished(HttpGrabber *, const QString &, void *, bool, bool)), this, SLOT(downloadFinished(HttpGrabber *, const QString &, void *, bool, bool)));
        downloadGrabbers_.append(grabber);
    }

#ifndef QT4
    resultImageDataCache_.setAutoDelete(true);
#endif
    resultImageDataCache_.setMaxCost(8 * 1024 * 1024); // 8 MB downloaded image cache...

    scanForMissingCoverArt();
}

void CoverArtDownloader::imageFlowViewBoxRearrangeLayers()
{
    imageFlowView_->setGeometry(QRect(0, 3, imageFlowViewBox_->width(), imageFlowViewBox_->height() - 3 - 20));
    flowScrollBar_->setGeometry(QRect(5, imageFlowViewBox_->height() - 20, imageFlowViewBox_->width() - 10, 20));
}

void CoverArtDownloader::closeEvent(QCloseEvent *e)
{
    emit closing();
    e->accept();
}

void CoverArtDownloader::done(int r)
{
    emit closing();
    QDialog::done(r);
}

void CoverArtDownloader::showEvent(QShowEvent *)
{
#ifdef QTOPIA
    showMaximized();
#endif
    updateListView();
}

void CoverArtDownloader::resizeEvent(QResizeEvent *)
{
    updateListView();
}

bool CoverArtDownloader::eventFilter(QObject *o, QEvent *e)
{
    if (e->type() == QEvent::KeyPress)
    {
        QKeyEvent *ke = static_cast<QKeyEvent *>(e);

        if (ke->key() == Qt::Key_Left)
        {
            imageFlowView_->showPreviousItem();
            ke->accept();
            return true;
        }
        else if (ke->key() == Qt::Key_Right)
        {
            imageFlowView_->showNextItem();
            ke->accept();
            return true;
        }
        else
            return QDialog::eventFilter(o, e);
    }
    else
        return QDialog::eventFilter(o, e);
}

void CoverArtDownloader::abortOnlineSearching()
{
    searchGrabbers_.abortAll();
    searchingOnline_ = false;
    currentAlbumIndex_ = -1;
}

void CoverArtDownloader::accept()
{
    abortOnlineSearching();

    splitter_->setEnabled(false);
    imageFlowViewBox_->hide();
    startDownloadButton_->setEnabled(false);

    downloadList_.clear();

    AlbumListViewItem *alvi;

#ifdef QT4
    for (int i = 0; i < albumListItems_.count(); ++i)
    {
        alvi = albumListItems_.at(i);
        if (alvi->checkState(0) == Qt::Checked && alvi->coverArtSearchResults.count() > 0 && alvi->selectedResultIndex != -1)
            downloadList_.append(alvi->coverArtSearchResults.at(alvi->selectedResultIndex));
    }
#else
    for (alvi = albumListItems_.first(); alvi != 0; alvi = albumListItems_.next())
        if (alvi->isOn() && alvi->coverArtSearchResults.count() > 0 && alvi->selectedResultIndex != -1)
            downloadList_.append(alvi->coverArtSearchResults.at(alvi->selectedResultIndex));
#endif

    downloadGrabbers_.abortAll();
    currentResultIndex_ = -1;
    downloadingImageFiles_ = true;
    currentAlbumListViewItem_ = NULL;
    coverArtSearchResultsToDownload_ = &downloadList_;

#ifdef QT4
    progress_->setMaximum(downloadList_.count());
#else
    progress_->setTotalSteps(downloadList_.count());
#endif
    progress_->show();

    for (int i = 0; i < downloadGrabbers_.count(); ++i)
        downloadNextResultImageData();
}

void CoverArtDownloader::reject()
{
    DENTERMETHOD;
    searchGrabbers_.abortAll();
    downloadGrabbers_.abortAll();
    QDialog::reject();
    DEXITMETHOD;
}

void CoverArtDownloader::scanForMissingCoverArt()
{
    DENTERMETHOD;
    DTIMERINIT(timer);

#ifdef QT4
    QProgressDialog progress(tr("Please wait..."), tr("Cancel"), 0, 0);
    progress.setModal(true);
#else
    QProgressDialog progress(tr("Please wait..."), tr("Cancel"), 0, NULL, "progress", true);
#endif

    progress.setWindowTitle(tr("Scanning for missing cover art image files..."));

    progress.setMinimumDuration(0);
    progress.setFixedWidth(qMin((int)(qApp->desktop()->width() * 0.9), 640));
    progress.setAutoClose(false);
    progress.setAutoReset(false);
    progress.show();
    qApp->processEvents();

    QString query =
        "SELECT artist, album, media_location.location || media.filename FROM " + playList_->sourceTable() + " AS pl "
        "LEFT JOIN media ON pl.media_id = media.media_id "
        "LEFT JOIN media_location ON media.location_id = media_location.location_id "
        "GROUP BY album ORDER BY artist, track ASC;";

    sqlite3_stmt *vm;
    sqlite3_prepare_v2(playList_->mediaDatabase()->db(), query.toUtf8(), -1, &vm, 0);

    albumListItems_.clear();
    listView_->clear();

    QStringList artists;
    QStringList albums;
    QStringList filenames;

    while (sqlite3_step(vm) == SQLITE_ROW)
    {
        artists.append(QString::fromUtf8((const char *)sqlite3_column_text(vm, 0)));
        albums.append(QString::fromUtf8((const char *)sqlite3_column_text(vm, 1)));
        filenames.append(QString::fromUtf8((const char *)sqlite3_column_text(vm, 2)));
    }

    sqlite3_finalize(vm);

#ifdef QT4
    progress.setMaximum(albums.count());
#else
    progress.setTotalSteps(albums.count());
#endif

    QTime time;
    time.start();

    for (int i = 0; i < albums.count(); ++i)
    {
        QString artist = artists[i];
        QString album = albums[i];
        QStringList imageFilenames = gCoverArtProvider.findCoverArtImageFilenames(artist, album, filenames[i]);

        if (album.length() > 0 && imageFilenames.count() == 0)
        {
            AlbumListViewItem *albumListViewItem = new AlbumListViewItem(listView_);
            albumListViewItem->setText(0, artist);
            albumListViewItem->setText(1, album);
            albumListViewItem->setText(2, "0");
            albumListViewItem->setText(3, "");
            albumListViewItem->setEnabled(false);
            albumListViewItem->setOn(false);
            albumListItems_.append(albumListViewItem);
        }

        if (time.elapsed() > 1000)
        {
#ifdef QT4
            progress.setValue(i);
#else
            progress.setProgress(i);
#endif
            progress.setLabelText(artist + " - " + album);
            qApp->processEvents();
            time.restart();
        }

#ifdef QT4
        if (progress.wasCanceled())
            break;
#else
        if (progress.wasCancelled())
            break;
#endif
    }

    DTIMERPRINT(timer, "Getting albums from DB...");

    DPRINTF("Album count: %d", albumListItems_.count());

    currentAlbumIndex_ = -1;
    searchingOnline_ = true;

    for (int i = 0; i < searchGrabbers_.count(); ++i)
        requestNextCoverArt();

    updateListView();

    DEXITMETHOD;
}

void CoverArtDownloader::updateListView()
{
#ifdef QT4
    listView_->resizeColumnToContents(2);
    listView_->resizeColumnToContents(3);
    listView_->resizeColumnToContents(4);

    int left = listView_->viewport()->width();
    left -= listView_->columnWidth(2);
    left -= listView_->columnWidth(3);
    left -= listView_->columnWidth(4);

    listView_->setColumnWidth(0, left / 2);
    listView_->setColumnWidth(1, left / 2);
    //listView_->setColumnWidth(5, 0);
#endif
}

void CoverArtDownloader::updateProgressBar()
{
    if (!searchingOnline_ && !downloadingImageFiles_)
    {
        progress_->hide();
        progressLabel_->hide();
    }
    else if (searchingOnline_)
    {
        progressLabel_->setText(tr("Searching online..."));

#ifdef QT4
        progress_->setMaximum(albumListItems_.count());
        progress_->setValue(currentAlbumIndex_);
#else
        progress_->setTotalSteps(albumListItems_.count());
        progress_->setProgress(currentAlbumIndex_);
#endif

        if (!progress_->isVisible())
            progress_->show();

        if (!progressLabel_->isVisible())
            progressLabel_->show();
    }
    else if (downloadingImageFiles_)
    {
        progressLabel_->setText(tr("Loading cover images..."));
#ifdef QT4
        progress_->setMaximum(coverArtSearchResultsToDownload_->count());
        progress_->setValue(currentResultIndex_);
#else
        progress_->setTotalSteps(coverArtSearchResultsToDownload_->count());
        progress_->setProgress(currentResultIndex_);
#endif

        if (!progress_->isVisible())
            progress_->show();

        if (!progressLabel_->isVisible())
            progressLabel_->show();
    }
}

void CoverArtDownloader::requestNextCoverArt()
{
    DENTERMETHOD;

    ++currentAlbumIndex_;
    updateProgressBar();

    if (searchingOnline_ && currentAlbumIndex_ < albumListItems_.count())
    {
        AlbumListViewItem *albumListViewItem = albumListItems_.at(currentAlbumIndex_);

        QString artist = CoverArtProvider::simplifyString(albumListViewItem->artist());
        QString album = CoverArtProvider::simplifyString(albumListViewItem->album());

#ifdef QT4
        artist = QUrl::toPercentEncoding(artist);
        album = QUrl::toPercentEncoding(album);
#else
        QUrl::encode(artist);
        QUrl::encode(album);
#endif

        QChar connectorChar = (addOn.coverArtWebServiceURL().contains('?') > 0 ? '&' : '?');
        QString requestURL = addOn.coverArtWebServiceURL() + connectorChar + "artist=" + artist + "&album=" + album;

        albumListViewItem->setText(4, "Searching...");

        HttpGrabber *grabber = searchGrabbers_.getIdleGrabber();
        grabber->download(requestURL, albumListViewItem);
        DPRINTF("Starting grabber %p for %s - %s, %s", grabber, (const char *)albumListViewItem->text(0).toUtf8(), (const char *)albumListViewItem->text(1).toUtf8(), (const char *)requestURL.toUtf8());
    }
    else
    {
        DPRINTF("Nothing to be done...");

        if (searchGrabbers_.busyGrabbersCount() == 0)
        {
            DPRINTF("Last grabber, finishing...");
            // FINISHED!
            currentAlbumIndex_ = -1;
            searchingOnline_ = false;
            updateProgressBar();
        }
    }

    DEXITMETHOD;
}


void CoverArtDownloader::searchFinished(HttpGrabber *grabber, const QString &url, void *userData, bool error, bool timedOut)
{
    DENTERMETHODF("[%p, %s, %p, %d, %d]", grabber, (const char *)url.toUtf8(), userData, error, timedOut);

    AlbumListViewItem *albumListViewItem = static_cast<AlbumListViewItem *>(userData);
    if (!error)
        albumListViewItem->setText(4, "");
    else
    {
        if (timedOut)
            albumListViewItem->setText(4, "Timed out");
        else
            albumListViewItem->setText(4, "Error");
    }

    requestNextCoverArt();
    DEXITMETHODF("[%p, %s, %p, %d, %d]", grabber, (const char *)url.toUtf8(), userData, error, timedOut);
}

void CoverArtDownloader::searchDataAvailable(HttpGrabber *grabber, const QString &url, void *userData, QByteArray *data)
{
    DENTERMETHODF("[%p, %s, %p, %p]", grabber, (const char *)url.toUtf8(), userData, data);

    AlbumListViewItem *albumListViewItem = static_cast<AlbumListViewItem *>(userData);
    albumListViewItem->coverArtSearchResults.clear();

    QString response(*data);

#ifdef QT4
    QStringList responseLines(response.split('\n', QString::SkipEmptyParts));
#else
    QStringList responseLines(QStringList::split('\n', response));
#endif

    for (QStringList::Iterator line = responseLines.begin(); line != responseLines.end(); ++line)
    {
#ifdef QT4
        QStringList lineItems((*line).split(":|:", QString::KeepEmptyParts));
#else
        QStringList lineItems(QStringList::split(":|:", (*line), true));
#endif

        DPRINTF("lineItems.count() = %d", lineItems.count());

        if (lineItems.count() == 4)
        {
            CoverArtSearchResult *item = new CoverArtSearchResult();
            item->title = lineItems[0];
            item->detailsURL = lineItems[1];
            item->thumbnailImageURL = lineItems[2];
            item->imageURL = lineItems[3];

            item->owner = albumListViewItem;

            albumListViewItem->coverArtSearchResults.append(item);
        }

        DPRINTF("%s \n", (const char *)(*line).toLatin1());
    }

    if (albumListViewItem->coverArtSearchResults.count() == 0)
    {
        albumListViewItem->setText(2, "0");
        albumListViewItem->setText(3, "");
        albumListViewItem->setEnabled(false);
        albumListViewItem->setOn(false);
    }
    else
    {
        albumListViewItem->setText(2, QString::number(albumListViewItem->coverArtSearchResults.count()));
        albumListViewItem->setText(3, "1");
        albumListViewItem->setEnabled(true);
        albumListViewItem->setOn(true);
    }

    if (responseLines.count() == 0)
    {
        DPRINTF("No results.");
    }

    DEXITMETHODF("[%p, %s, %p, %p]", grabber, (const char *)url.toUtf8(), userData, data);
}

void CoverArtDownloader::listViewCurrentItemChanged()
{
    loadFlowView(static_cast<AlbumListViewItem *>(listView_->currentItem()));
}

void CoverArtDownloader::loadFlowView(AlbumListViewItem *alvi)
{
    imageFlowView_->images().clear();
    downloadGrabbers_.abortAll();

    for (int i = 0; i < alvi->coverArtSearchResults.count(); ++i)
    {
        QImage *image = new QImage();
        imageFlowView_->images().append(image);
    }

    imageFlowView_->setAlbumListViewItem(alvi);

    currentResultIndex_ = -1;
    downloadingImageFiles_ = true;
    currentAlbumListViewItem_ = alvi;
    coverArtSearchResultsToDownload_ = &alvi->coverArtSearchResults;

    imageFlowViewBoxRearrangeLayers();

    for (int i = 0; i < downloadGrabbers_.count(); ++i)
        downloadNextResultImageData();
}

void CoverArtDownloader::downloadNextResultImageData()
{
    DENTERMETHOD;

    ++currentResultIndex_;
    updateProgressBar();

    if (downloadingImageFiles_ && coverArtSearchResultsToDownload_ && currentResultIndex_ < coverArtSearchResultsToDownload_->count())
    {
        CoverArtSearchResult *coverArtSearchResult = coverArtSearchResultsToDownload_->at(currentResultIndex_);
#ifdef QT4
        QByteArray *cachedImageData = resultImageDataCache_.object(coverArtSearchResult->imageURL);
#else
        QByteArray *cachedImageData = resultImageDataCache_.find(coverArtSearchResult->imageURL, true);
#endif

        if (cachedImageData)
        {
            handleCoverArtSearchResultImageData(coverArtSearchResult, cachedImageData);
            downloadNextResultImageData();
        }
        else
        {
            HttpGrabber *grabber = downloadGrabbers_.getIdleGrabber(); // there has to be at least one idle grabber in our case...
            grabber->download(coverArtSearchResult->imageURL, coverArtSearchResult);

            DPRINTF("Starting grabber %p for download of %s, %s - %s",
                grabber, (const char *)coverArtSearchResult->imageURL.toUtf8(),
                (const char *)coverArtSearchResult->owner->artist().toUtf8(), (const char *)coverArtSearchResult->owner->album().toUtf8());
        }
    }
    else
    {
        DPRINTF("Nothing to be done...");

        if (downloadGrabbers_.busyGrabbersCount() == 0)
        {
            DPRINTF("Last grabber, finishing...");

            currentResultIndex_ = -1;
            downloadingImageFiles_ = false;
            coverArtSearchResultsToDownload_ = NULL;

            updateProgressBar();

            if (!currentAlbumListViewItem_)
                QDialog::accept();
        }
    }

    DEXITMETHOD;
}

void CoverArtDownloader::handleCoverArtSearchResultImageData(CoverArtSearchResult *coverArtSearchResult, QByteArray *imageData)
{
    if (currentAlbumListViewItem_)
    {
        if (coverArtSearchResult->owner == currentAlbumListViewItem_)
        {
#ifdef QT4
            int index = currentAlbumListViewItem_->coverArtSearchResults.indexOf(coverArtSearchResult);
#else
            int index = currentAlbumListViewItem_->coverArtSearchResults.findRef(coverArtSearchResult);
#endif

            // prepare image
            QImage *image = imageFlowView_->images().at(index);
            image->loadFromData(*imageData);
            imageFlowView_->invalidate();
        }
    }
    else
    {
#ifdef QT4
        if (!QDir(qConfig.coverArtDirectory).exists())
            QDir().mkpath(qConfig.coverArtDirectory);
#else
        if (!QDir(qConfig.coverArtDirectory).exists())
            QDir().mkdir(qConfig.coverArtDirectory, true);
#endif

        QString coverArtDirectory = qConfig.coverArtDirectory + QDir::separator() + "00DownloadedCoverArt";

#ifdef QT4
        if (!QDir(coverArtDirectory).exists())
            QDir().mkpath(coverArtDirectory);
#else
        if (!QDir(coverArtDirectory).exists())
            QDir().mkdir(coverArtDirectory, true);
#endif

        QString filename = coverArtDirectory + QDir::separator() +
            CoverArtProvider::simplifyString(coverArtSearchResult->owner->artist()) + " - " +
            CoverArtProvider::simplifyString(coverArtSearchResult->owner->album()) + ".jpg";

        QFile f(filename);
#ifdef QT4
        f.open(QIODevice::WriteOnly);
        f.write(*imageData);
#else
        f.open(IO_WriteOnly);
        f.writeBlock(*imageData);
#endif
        f.close();
    }
}

void CoverArtDownloader::downloadFinished(HttpGrabber *grabber, const QString &url, void *userData, bool error, bool timedOut)
{
    DENTERMETHODF("[%p, %s, %p, %d, %d]", grabber, (const char *)url.toUtf8(), userData, error, timedOut);
    CoverArtSearchResult *coverArtSearchResult = static_cast<CoverArtSearchResult *>(userData);
    if (!error)
        coverArtSearchResult->owner->setText(4, "OK");
    else
    {
        if (timedOut)
            coverArtSearchResult->owner->setText(4, "Timed out");
        else
            coverArtSearchResult->owner->setText(4, "Error");
    }
#ifdef QT4
    // TODO: coverArtSearchResult->owner->treeWidget()->scrollTo( coverArtSearchResult->owner, QAbstractItemView::EnsureVisible);
#else
    coverArtSearchResult->owner->listView()->ensureItemVisible(coverArtSearchResult->owner);
#endif

    downloadNextResultImageData();
    DEXITMETHODF("[%p, %s, %p, %d, %d]", grabber, (const char *)url.toUtf8(), userData, error, timedOut);
}

void CoverArtDownloader::downloadDataAvailable(HttpGrabber *grabber, const QString &url, void *userData, QByteArray *data)
{
    DENTERMETHODF("[%p, %s, %p, %p]", grabber, (const char *)url.toUtf8(), userData, data);
    CoverArtSearchResult *coverArtSearchResult = static_cast<CoverArtSearchResult *>(userData);
    QByteArray *result = new QByteArray(*data);
    resultImageDataCache_.insert(coverArtSearchResult->thumbnailImageURL, result, result->size());
    handleCoverArtSearchResultImageData(coverArtSearchResult, result);
    DEXITMETHODF("[%p, %s, %p, %p]", grabber, (const char *)url.toUtf8(), userData, data);
}
