#ifndef COVERARTDOWNLOADER_H_
#define COVERARTDOWNLOADER_H_

#include <qwidget.h>
#include <qlayout.h>
#include <qdialog.h>
#ifdef QT4
#include <QHttp>
#include <QHeaderView>
#include <QTreeWidget>
#include <QTreeWidgetItem>
#include <QScrollBar>
#include <QMenuBar>
#include <QMenu>
#include <QGroupBox>
#include <QSplitter>
#else
#include <qhttp.h>
#include <qlistview.h>
#include <qhgroupbox.h>
#include <qsplitter.h>
#endif
#include <qlist.h>
#include <qcache.h>
#include <qprogressbar.h>
#include <qmap.h>
#include <qlabel.h>

#include "addonmanager.h"
#include "configurationdialog.h"
#include "coverartconfig.h"
#include <qlineedit.h>

#include "paintbox.h"
#include "scrollbar.h"
#include "customperspectiveimageflow.h"

#include "quasar.h"
#include "configuration.h"
#include "helpers.h"
#include "playlist.h"
#include "skin.h"
#include "skinmanager.h"

#define COVERART_DEFAULT_URL "http://coverart.katastrophos.net/query.php"

class QSplitter;
class AlbumListViewItem;

class CoverArtSearchResult
{
public:
    QString title;
    QString detailsURL;
    QString thumbnailImageURL;
    QString imageURL;

    AlbumListViewItem *owner;
};

#ifdef QT4
typedef QList<CoverArtSearchResult *> CoverArtSearchResultList;
class AlbumListViewItem : public QTreeWidgetItem
{
public:
    AlbumListViewItem(QTreeWidget *parent)
        :   QTreeWidgetItem(parent),
            selectedResultIndex(0)
    {
    }

    ~AlbumListViewItem()
    {
        qDeleteAll(coverArtSearchResults);
    }

    QString artist() const { return text(0); }
    QString album() const { return text(1); }

    bool isOn() { return checkState(0) == Qt::Checked; }

    void setEnabled(bool value) { setDisabled(!value); }
    void setOn(bool value) { setCheckState(0, value ? Qt::Checked : Qt::Unchecked); }

    CoverArtSearchResultList coverArtSearchResults;
    int selectedResultIndex;
};

typedef QList<AlbumListViewItem *> AlbumListViewItemList;
#else
typedef QList<CoverArtSearchResult> CoverArtSearchResultList;
class AlbumListViewItem : public QCheckListItem
{
public:
    AlbumListViewItem(QListView *parent)
        :   QCheckListItem(parent, "", CheckBox),
            selectedResultIndex(0)
    {
        coverArtSearchResults.setAutoDelete(true);
    }

    QString artist() const { return text(0); }
    QString album() const { return text(1); }

    CoverArtSearchResultList coverArtSearchResults;
    int selectedResultIndex;
};

typedef QList<AlbumListViewItem> AlbumListViewItemList;
#endif

class HttpGrabber : public QObject
{
    Q_OBJECT
public:
    HttpGrabber(QObject *parent = 0);
    virtual ~HttpGrabber();

    void setTimeout(uint msec) { timeout_ = msec; }
    uint timeout() { return timeout_; }
    bool download(const QString &url, void *userData = NULL);
    bool isBusy() { return busy_; };
    void abort();

signals:
    void finished(HttpGrabber *, const QString &url, void *userData, bool error, bool timedOut);
    void dataAvailable(HttpGrabber *, const QString &url, void *userData, QByteArray *data);

private slots:
    void timerTimeout();
    void httpRequestFinished(int res, bool error);

private:
    QTimer timer_;
    QHttp http_;
    QString url_;
    void *userData_;
    uint timeout_;
    bool busy_;
    int downloadRes_;
    uint lastBytesAvailable_;
};

#ifdef QT4
typedef QList<HttpGrabber *> HttpGrabberListPrototype;
#else
typedef QList<HttpGrabber> HttpGrabberListPrototype;
#endif

class HttpGrabberList : public HttpGrabberListPrototype
{
public:
#ifdef QT4
    ~HttpGrabberList()
    {
        qDeleteAll(*this);
    }
#endif

    void abortAll()
    {
        for (int i = 0; i < count(); ++i)
            at(i)->abort();
    }

    int busyGrabbersCount()
    {
        int result = 0;
        for (int i = 0; i < count(); ++i)
            if (at(i)->isBusy())
                ++result;

        return result;
    }

    HttpGrabber *getIdleGrabber()
    {
        HttpGrabber *g = NULL;

        for (int i = 0; i < count(); ++i)
            if (!at(i)->isBusy())
            {
                g = at(i);
                break;
            }

        return g;
    }
};

#ifdef QT4
typedef QList<QImage *> QImageList;
#else
typedef QList<QImage> QImageList;
#endif

class DownloadedCoverArtImageFlow : public CustomPerspectiveImageFlow
{
    Q_OBJECT
public:
    DownloadedCoverArtImageFlow(LayeredPaintBox *parent);
    ~DownloadedCoverArtImageFlow();
    QImageList &images() { return images_; }
    void invalidate() { CustomPerspectiveImageFlow::invalidate(); }
    void setAlbumListViewItem(AlbumListViewItem *albumListViewItem);

protected:
    QImage getImage(int index);
    void paintOverlay(QPainter &p, const QPixmap &pixmap);
    void transitionFinished();

private:
    QImageList images_;
    AlbumListViewItem *albumListViewItem_;
};

class CoverArtDownloader : public QDialog
{
    Q_OBJECT
public:
    CoverArtDownloader(PlayList* playList, SkinManager* skinManager, QWidget *parent = 0);

#ifdef QT4
    QTreeWidget *listView() { return listView_; }
    QSplitter *splitter() { return splitter_; }
#else
    QListView *listView() { return listView_; }
    QSplitter *splitter() { return splitter_; }
#endif

signals:
    void closing();

protected slots:
    virtual void accept();
    virtual void reject();

protected:
    void closeEvent(QCloseEvent *);
    void done(int r);
    void showEvent(QShowEvent *);
    void resizeEvent(QResizeEvent *);

private slots:
    void imageFlowViewBoxRearrangeLayers();

    void scanForMissingCoverArt();

    void listViewCurrentItemChanged();

    void loadFlowView(AlbumListViewItem *);

    void searchFinished(HttpGrabber *grabber, const QString &url, void *userData, bool error, bool timedOut);
    void searchDataAvailable(HttpGrabber *grabber, const QString &url, void *userData, QByteArray *data);

    void downloadFinished(HttpGrabber *grabber, const QString &url, void *userData, bool error, bool timedOut);
    void downloadDataAvailable(HttpGrabber *grabber, const QString &url, void *userData, QByteArray *data);

private:
    PlayList *playList_;
    SkinManager* skinManager_;
#ifdef QT4
    QTreeWidget *listView_;
#else
    QListView *listView_;
#endif
    QLabel *progressLabel_;
    QSplitter *splitter_;
    QProgressBar *progress_;
    LayeredPaintBox *imageFlowViewBox_;
    DownloadedCoverArtImageFlow *imageFlowView_;
    FlowScrollBar *flowScrollBar_;
    QPushButton *startDownloadButton_;
    QPushButton *cancelButton_;

    AlbumListViewItemList albumListItems_;

    int currentAlbumIndex_;
    bool searchingOnline_;
    HttpGrabberList searchGrabbers_;

    AlbumListViewItem *currentAlbumListViewItem_;
    CoverArtSearchResultList *coverArtSearchResultsToDownload_;
    CoverArtSearchResultList downloadList_;
    int currentResultIndex_;
    bool downloadingImageFiles_;

    HttpGrabberList downloadGrabbers_;

#ifdef QT4
    QCache<QString, QByteArray> resultImageDataCache_;
#else
    QCache<QByteArray> resultImageDataCache_;
#endif

    void handleCoverArtSearchResultImageData(CoverArtSearchResult *coverArtSearchResult, QByteArray *imageData);
    void requestNextCoverArt();
    void abortOnlineSearching();
    void updateProgressBar();
    void updateListView();
    void downloadNextResultImageData();

    bool eventFilter(QObject *object, QEvent *e);
};


class CoverArtDownloaderAddOn : public AddOn
{
    Q_OBJECT
public:
    CoverArtDownloaderAddOn()
        :	AddOn(),
            coverArtWebServiceURL_(COVERART_DEFAULT_URL),
            dialog_(0)
    {
        qDebug("Registering CoverArtDownloaderAddOn");
    }

    virtual ~CoverArtDownloaderAddOn()
    {
    }

    virtual void hookMainWindow(Quasar *mainWindow)
    {
        mainWindow_ = mainWindow;

#ifdef QT4
        QAction *newEntry = mainWindow->toolsMenu_->addAction(tr("Cover Art Downloader..."), this, SLOT(openCoverArtDownloader()));
        mainWindow->toolsMenu_->insertAction(mainWindow_->idToolsMenuRescanForCoverArt_, newEntry);
        if (mainWindow->mobileMenuBar_)
            mainWindow->mobileMenuBar_->addAction(newEntry);
#else
        int insertIndex = mainWindow->toolsMenu_->indexOf(mainWindow_->idToolsMenuRescanForCoverArt_) + 1;
        qDebug("insertIndex: %d", insertIndex);
        mainWindow->toolsMenu_->insertItem(tr("Cover Art Downloader..."), this, SLOT(openCoverArtDownloader()), 0, -1, insertIndex);
#endif
    }

    virtual void hookConfigurationDialog(ConfigurationDialog *cd)
    {
        cd_ = cd;

#ifdef QT4
        QGroupBox *coverArtDownloaderGroup = new QGroupBox(tr("Cover Art Download Settings"), cd_->coverArtConfig->widget());
        QHBoxLayout *layout = new QHBoxLayout(coverArtDownloaderGroup);
        cd_->coverArtConfig->widget()->layout()->addWidget(coverArtDownloaderGroup);
#else
        QHGroupBox *coverArtDownloaderGroup = new QHGroupBox(tr("Cover Art Download Settings"), cd_->coverArtConfig->widget());
        cd_->coverArtConfig->widget()->layout()->add(coverArtDownloaderGroup);
#endif

        coverArtDownloadWebServiceURL_ = new QLineEdit(coverArtDownloaderGroup);
#ifdef QT4
        layout->addWidget(coverArtDownloadWebServiceURL_);
#endif
        connect(coverArtDownloadWebServiceURL_, SIGNAL(textChanged(const QString&)),
                this, SLOT(setCoverArtDownloadWebServiceURL(const QString&)));

        connect(cd_, SIGNAL(settingStatesByConfiguration(const Configuration&)),
                this, SLOT(setStatesByConfiguration(const Configuration&)));

        connect(cd_, SIGNAL(settingDefaultSettings()),
                this, SLOT(setDefaultSettings()));
    }

    virtual void writeToConfig(Config &cf)
    {
        cf.setGroup("CoverArt");
        cf.writeEntry("coverArtWebServiceURL", coverArtWebServiceURL_);
    }

    virtual void readFromConfig(Config &cf)
    {
        cf.setGroup("CoverArt");
        coverArtWebServiceURL_ = cf.readEntry("coverArtWebServiceURL", coverArtWebServiceURL_);
    }

    const QString &coverArtWebServiceURL() { return coverArtWebServiceURL_; }

private slots:
    void openCoverArtDownloader()
    {
        if (!dialog_)
        {
            dialog_ = new CoverArtDownloader(&mainWindow_->playListManager_->currentPlayList(), mainWindow_->skinManager_, mainWindow_);

#ifdef MAEMO
            setStyleRecursively(dialog_, mainWindow_->style());
            dialog_->splitter()->setPalette(mainWindow_->horizontalSplitterContainer_->palette());
            dialog_->listView()->setPalette(mainWindow_->playListView_->palette());
            dialog_->listView()->setFont(mainWindow_->playListView_->font());
            dialog_->listView()->header()->setPalette(mainWindow_->playListView_->header()->palette());
            dialog_->listView()->header()->setFont(mainWindow_->playListView_->header()->font());
            dialog_->listView()->horizontalScrollBar()->setPalette(mainWindow_->playListView_->horizontalScrollBar()->palette());
            dialog_->listView()->verticalScrollBar()->setPalette(mainWindow_->playListView_->verticalScrollBar()->palette());
#endif
            connect(dialog_, SIGNAL(closing()), this, SLOT(closeCoverArtDownloader()));
            loadWidgetPosition(dialog_, qGlobalConfig.configFilename(), "CoverArtDownloader");
        }

#ifdef QTOPIA
        dialog_->showMaximized();
#else
        dialog_->show();
#ifdef QT4
        dialog_->setWindowState(dialog_->windowState() & ~Qt::WindowMinimized | Qt::WindowActive);
#else
        dialog_->setWindowState(dialog_->windowState() & ~WindowMinimized | WindowActive);
#endif
#endif
    }

    void closeCoverArtDownloader()
    {
        mainWindow_->rescanForCoverArt();
        saveWidgetPosition(dialog_, qGlobalConfig.configFilename(), "CoverArtDownloader");
        dialog_ = NULL;
    }

    void setCoverArtDownloadWebServiceURL(const QString& url)
    {
        coverArtWebServiceURL_ = url;
    }

    void setStatesByConfiguration(const Configuration& config)
    {
        coverArtDownloadWebServiceURL_->setText(coverArtWebServiceURL_);
    }

    void setDefaultSettings()
    {
        coverArtDownloadWebServiceURL_->setText(COVERART_DEFAULT_URL);
    }

private:
    Quasar *mainWindow_;

    ConfigurationDialog *cd_;
    QLineEdit *coverArtDownloadWebServiceURL_;

    CoverArtDownloader *dialog_;

    QString coverArtWebServiceURL_;
};

#endif /*COVERARTDOWNLOADER_H_*/
